# -*- shell-script -*-

dnl Some local m4 macros for use in configure.in

############################################################################

define([LSC_CHECK_MPI],[
#
# Try to link a simple program; requires both <mpi.h> and
# MPI_Initialized.
# 

AC_MSG_CHECKING([for MPI])
AC_TRY_LINK(#include <mpi.h>
, int i; MPI_Initialized(&i);, LSC_MPI_FOUND=yes, LSC_MPI_FOUND=no)
if test "$LSC_MPI_FOUND" = "no"; then
    LIBS_save="$LIBS"
    LIBS="$LIBS -lmpi"
    AC_TRY_LINK(#include <mpi.h>
, int i; MPI_Initialized(&i);, LSC_MPI_FOUND=yes, LSC_MPI_FOUND=no)
    if test "$LSC_MPI_FOUND" = "no"; then
	LIBS="$LIBS_save"
    fi
fi
AC_MSG_RESULT([$LSC_MPI_FOUND])])dnl

############################################################################

define([LSC_CHECK_MPI_DEPDIRS],[
#
# Looks for the include directories necessary for MPI.
# Will call AC_PROG_CPP if it has not been called alreay.
#
# Sets LSC_MPI_DEPDIRS to "-Idir1 -Idir2..." or blank
# Calls AC_SUBST on LSC_MPI_DEPDIRS
#
# Necessary for finding dependencies (e.g., <mpi.h>) in "make depend"
# rules.
#

# Now see if we've figured out what the C preprocessor is
if test "$?CPP" = "0"; then
    AC_PROG_CPP
elif test "$CPP" = ""; then
    AC_PROG_CPP
fi
AC_MSG_CHECKING([for MPI include directories])
cat > conftest.c <<EOF
#include "mpi.h"

int main(int argc, char* argv[]) { return 0; }
EOF
echo configure:__oline__: $CPP $CPPFLAGS conftest.C >&5
$CPP $CPPFLAGS conftest.c | grep '^#.*\"' | grep -v '^#pragma' | \
grep -v '^#ident' | sed -e 's/#.*\"\(.*\)\".*$/\1/' | \
sort -u - | grep -v conftest.c > conftest.out 2>&1
LSC_TEMP=
rm -f conftest2.out
touch conftest2.out
for LSC_FILE in `cat conftest.out`; do
    LSC_TEMP2="`dirname $LSC_FILE`"
    if test "$LSC_TEMP" != "$LSC_TEMP2"; then
	echo "$LSC_TEMP2" >> conftest2.out
	LSC_TEMP="$LSC_TEMP2"
    fi
done
LSC_MPI_DEPDIRS=
for LSC_DIR in `sort -u conftest2.out | uniq`; do
	if test -f $LSC_DIR/mpi.h ; then
    	LSC_MPI_DEPDIRS="-I$LSC_DIR $LSC_MPI_DEPDIRS"
	fi
done
rm -f confest*
AC_MSG_RESULT([$LSC_MPI_DEPDIRS])
AC_SUBST(LSC_MPI_DEPDIRS)])dnl

############################################################################

define([LSC_CHECK_MPI_COMM_GET_PARENT],[
#
# Checks for MPI_Comm_get_parent (it's an MPI-2 function, not all MPI
# implementations have it yet)
#

AC_MSG_CHECKING([for MPI_Comm_get_parent])
AC_TRY_LINK(#include <mpi.h>
, MPI_Comm foo; MPI_Comm_get_parent(&foo);, 
LSC_MPI_COMM_GET_PARENT=yes, LSC_MPI_COMM_GET_PARENT=no)
if test "$LSC_MPI_COMM_GET_PARENT" = "yes"; then
    LSC_HAVE_MPI_COMM_GET_PARENT=1
else
    LSC_HAVE_MPI_COMM_GET_PARENT=0
fi
AC_MSG_RESULT([$LSC_MPI_COMM_GET_PARENT])])dnl

############################################################################

define([LSC_CHECK_MPI_COMM_SPAWN],[
#
# Checks for MPI_Comm_get_parent (it's an MPI-2 function, not all MPI
# implementations have it yet)
#

AC_MSG_CHECKING([for MPI_Comm_spawn])
AC_TRY_LINK(#include <mpi.h>
, char **argv = (char**) 0; 
MPI_Info info; 
MPI_Comm intercomm;
int errs;
MPI_Comm_spawn("foo", argv, 1, info, 0, MPI_COMM_WORLD, &intercomm, &errs);, 
LSC_MPI_COMM_SPAWN=yes, LSC_MPI_COMM_SPAWN=no)
if test "$LSC_MPI_COMM_SPAWN" = "yes"; then
    LSC_HAVE_MPI_COMM_SPAWN=1
else
    LSC_HAVE_MPI_COMM_SPAWN=0
fi
AC_MSG_RESULT([$LSC_MPI_COMM_SPAWN])])dnl

############################################################################

define([LSC_CHECK_MPI_STATUS_IGNORE],[
#
# See if we have MPI_STATUS_IGNORE
# 

AC_MSG_CHECKING([for MPI_STATUS_IGNORE])
AC_TRY_LINK(#include <mpi.h>
, int i = sizeof(MPI_STATUS_IGNORE);, 
LSC_MPI_STATUS_IGNORE=yes, LSC_MPI_STATUS_IGNORE=no)
if test "$LSC_MPI_STATUS_IGNORE" = "no"; then
    LSC_HAVE_MPI_STATUS_IGNORE=0
else
    LSC_HAVE_MPI_STATUS_IGNORE=1
fi
AC_MSG_RESULT([$LSC_MPI_STATUS_IGNORE])])dnl

############################################################################

define([LSC_CHECK_MPI_COMM_SET_NAME],[
#
# See if we have MPI_COMM_SET_NAME
# 

AC_MSG_CHECKING([for MPI_Comm_set_name])
AC_TRY_LINK(#include <mpi.h>
, void *foo = (void*) MPI_Comm_set_name;, 
LSC_MPI_COMM_SET_NAME=yes, LSC_MPI_COMM_SET_NAME=no)
if test "$LSC_MPI_COMM_SET_NAME" = "no"; then
    LSC_HAVE_MPI_COMM_SET_NAME=0
else
    LSC_HAVE_MPI_COMM_SET_NAME=1
fi
AC_MSG_RESULT([$LSC_MPI_COMM_SET_NAME])])dnl

