/*
			(c) Copyright 1998, 1999 - Tord Jansson
			=======================================

		This file is part of the BladeEnc MP3 Encoder, based on
		ISO's reference code for MPEG Layer 3 compression.

		This file doesn't contain any of the ISO reference code and
		is copyright Tord Jansson (tord.jansson@swipnet.se).

	BladeEnc is free software; you can redistribute this file
	and/or modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "system.h"
#include "codec.h"
#include "samplein.h"
#include "arglink.h"
#include "bladesys.h"
#include "parallel.h"
#include "memdeb.h"
#if WANT_MPI
#include <mpi.h>
#endif

#ifdef OS2
#	define INCL_DOSFILEMGR
#	define INCL_DOSERRORS
#	define INCL_DOSPROCESS
#	include <os2.h>
#	include "os2key.h"
#endif

#if SYSTEM == MAC_OS
#	include <SIOUX.h>
#	include <Events.h>

void DoEvent (EventRecord * i);
extern int gFriendliness;
#endif
/*
void updateProgressIndicator (time_t startBatch, double batchLen,
			      double batchDone, time_t startFile,
			      double fileLen, double fileDone, int progType,
			      Job * psJob);*/
void					updateProgressIndicator
(
	time_t					startBatch,
	double					batchLen,
	double					batchDone,
	time_t					startFile,
	double					fileLen,
	double					fileDone, 
	time_t					currTime,
	int						progType,
	Job						*psJob
);

/*____ External Data __________________________________________________________*/

/*extern int wantedProgressBarRate;*/
extern int wantedQuit;
extern int wantedQuiet;
extern Job *psJobQueue;
extern FILE *textout;
extern int progType;


/*____ Static Data ____________________________________________________________*/

#if WANT_MPI
static MPI_Datatype codecInitInDatatype = MPI_DATATYPE_NULL;
static MPI_Datatype splInDatatype = MPI_DATATYPE_NULL;
static MPI_Datatype jobDatatype = MPI_DATATYPE_NULL;
static MPI_Datatype slaveJobDatatype = MPI_DATATYPE_NULL;
static MPI_Datatype slaveReturnDatatype = MPI_DATATYPE_NULL;

static int last_written = -1;
static queue_entry *head = NULL;
#endif /* WANT_MPI */

/*____ Local functions ________________________________________________________*/

#if WANT_MPI
static void reset_result_queue(void);
static void queue_result(FILE *fp, char *data, slave_return *ret, int index);
#endif /* WANT_MPI */


/*____ serial_master() ________________________________________________________*/

/* Called when we don't compile with parallelism, or if there is only
   one slave (i.e., this rank) available */
void
serial_master(int samplesPerFrame, char *pBuffer, FILE *fp,
	      time_t startTimeBatch, time_t startTimeFile,
	      double batchSamplesTotal, double *fileSamplesRead,
	      double *batchSamplesRead)
{
  int nSamples;
  int nSamples2;
  short readBuffer[2304];
  uint encodedChunkSize;
  int showProgressCnt = 0;
  int overflowBits;
  char input;
  int nNoGapSamples = 0;
  int noGapOfs = 0;
  time_t currTime, oldTime = 0;

		while  ((nSamples = readSamples (&psJobQueue->sInput, samplesPerFrame - nNoGapSamples, readBuffer + noGapOfs)) > 0)
		{
			if ((nNoGapSamples != 0) || (noGapOfs !=0))			
				printf("nNoGapSamples: %d , noGapOfs: %d  --- \n", nNoGapSamples, noGapOfs);			
			nSamples     += nNoGapSamples;
			nNoGapSamples = 0;
			noGapOfs      = 0;

			if (psJobQueue->bNoGap  &&  nSamples < samplesPerFrame)
				break;	

			/* Stuff needed for Petteri Kamppuri's Mac OS port */

#if SYSTEM == MAC_OS	
	 		if (macTickStart + gFriendliness < TickCount ())
			{
	        	macTickStart = TickCount ();
	
	       		Boolean					gotEvent;
	         	EventRecord				event;
	
	         	gotEvent = GetNextEvent (everyEvent, &event);
	         	if (gotEvent)
				{
	            	SIOUXHandleOneEvent (&event);
	           		DoEvent (&event);
	         	}
			}
#endif

			/* Encode a chunk and Write to stream */

			nSamples2 = nSamples;
			if (psJobQueue->sCodec.mode != 3) 			
				nSamples2 *= 2;
			
			encodedChunkSize = codecEncodeChunk (nSamples2, readBuffer, pBuffer);
			if (fwrite (pBuffer, 1, encodedChunkSize, fp) != encodedChunkSize)
			{
				fprintf (textout, "ERROR: Couldn't write '%s'! Disc probably full.\n", psJobQueue->outputFilename);
				quit (6);
			}


			/* Update progress indicator */

			*batchSamplesRead += nSamples;
			*fileSamplesRead  += nSamples;

			if (!wantedQuiet )                         /*&& progType != PROG_OFF)*/
			{
				currTime = time (NULL);
				if (currTime != oldTime)
				{
					oldTime = currTime;
					updateProgressIndicator (startTimeBatch, batchSamplesTotal, *batchSamplesRead,
					                         startTimeFile, psJobQueue->sInput.length, *fileSamplesRead, 
											 currTime, progType, psJobQueue);
				}
			}
			/*if (!wantedQuiet)
			{
				showProgressCnt = (showProgressCnt + 1) % wantedProgressBarRate;
				if (showProgressCnt == 0)
					updateProgressIndicator (startTimeBatch, batchSamplesTotal,
							   *batchSamplesRead, startTimeFile,
							   psJobQueue->sInput.length,
							   *fileSamplesRead, progType,
							   psJobQueue);
			}*/


			/* Handling pause through ESC (Windows & OS/2 only) */

			if (be_kbhit() != 0)
			{
				input = be_getch ();
				if (input == 27)
				{
					fprintf (textout, "\r                                                                             \r");
					fprintf (textout, "Quit, are you sure? (y/n)");
					fflush (textout);
					input = be_getch();
					if (input == 'y'  ||  input == 'Y')
					{
						encodedChunkSize = codecExit (pBuffer, &overflowBits);
						if (encodedChunkSize != 0)
							if (fwrite (pBuffer, encodedChunkSize, 1, fp) != 1)
							{
								fprintf (textout, "ERROR: Couldn't write '%s'! Disc probably full.\n", psJobQueue->outputFilename);
								quit (6);
							}
						free (pBuffer);
						closeInput (&psJobQueue->sInput);
						if (fp != stdout)
							fclose (fp);
						exit (0);
					}
					else
						fprintf (textout, "\r                                                                             \r");
				}

			}
		}
		*batchSamplesRead += nSamples;
		*fileSamplesRead += nSamples;

#if 0
  while ((nSamples = readSamples (&psJobQueue->sInput, samplesPerFrame,
				  readBuffer)) > 0) {
    
#if SYSTEM == MAC_OS		/* Stuff needed for Petteri Kamppuri's Mac OS port */
    if (macTickStart + gFriendliness < TickCount ())
      {
	macTickStart = TickCount ();
	
	Boolean gotEvent;
	EventRecord event;
	
	gotEvent = GetNextEvent (everyEvent, &event);
	if (gotEvent)
	  {
	    SIOUXHandleOneEvent (&event);
	    DoEvent (&event);
	  }
      }
#endif

    encodedChunkSize = codecEncodeChunk (nSamples, readBuffer, pBuffer);
    if (fwrite (pBuffer, 1, encodedChunkSize, fp) != encodedChunkSize)
      {
	fprintf (textout,
		 "ERROR: Couldn't write '%s'! Disc probably full.\n",
		 psJobQueue->outputFilename);
	quit (-1);
      }
    
    *batchSamplesRead += nSamples;
    *fileSamplesRead += nSamples;
    
    if (!wantedQuiet)
      {
	showProgressCnt = (showProgressCnt + 1) % wantedProgressBarRate;
	if (showProgressCnt == 0)
	  updateProgressIndicator (startTimeBatch, batchSamplesTotal,
				   *batchSamplesRead, startTimeFile,
				   psJobQueue->sInput.length,
				   *fileSamplesRead, progType,
				   psJobQueue);
      }
    if (be_kbhit () != 0)
      {
	input = be_getch ();
	if (input == 27)
	  {
	    fprintf (textout,
		     "\r                                                                             \r");
	    fprintf (textout, "Quit, are you sure? (y/n)");
	    fflush (textout);
	    input = be_getch ();
	    if (input == 'y' || input == 'Y')
	      {
		encodedChunkSize = codecExit (pBuffer, &overflowBits);
		if (encodedChunkSize != 0)
		  if (fwrite (pBuffer, encodedChunkSize, 1, fp) != 1)
		    {
		      fprintf (textout,
			       "ERROR: Couldn't write '%s'! Disc probably full.\n",
			       psJobQueue->outputFilename);
		      quit (-1);
		    }
		free (pBuffer);
		closeInput (&psJobQueue->sInput);
		if (fp != stdout)
		  fclose (fp);
		exit (0);
	      }
	    else
	      fprintf (textout,
		       "\r                                                                             \r");
	  }
      }
  }
#endif
}

#if WANT_MPI

/*____ make_mpi_datatypes() ___________________________________________________*/

void
make_mpi_datatypes(void)
{
  int i;

  /* codecInitIn */
  {
    MPI_Type_vector(8, 1, 1, MPI_INT, &codecInitInDatatype);
    MPI_Type_commit(&codecInitInDatatype);
  }

  /* splInDatatype */
  {
    /*SplIn s;*/
    SI_Stream s;
    int lengths[7] = { 2, 1, 2, 1, 7, 256, 1 };
    MPI_Aint disps[7];
    MPI_Datatype types[7] = { MPI_INT, MPI_UNSIGNED, MPI_INT, MPI_UNSIGNED,
			      MPI_INT, MPI_CHAR, MPI_INT };

    disps[0] = (MPI_Aint) &s.outputType;
    disps[1] = (MPI_Aint) &s.length;
    disps[2] = (MPI_Aint) &s.errcode;
    disps[3] = (MPI_Aint) &s.samplesLeft;
    disps[4] = (MPI_Aint) &s.freq;
    disps[5] = (MPI_Aint) &s.preReadBuffer;
    disps[6] = (MPI_Aint) &s.nPreReadBytes;
    for (i = 0; i < 7; i++)
      disps[i] -= (MPI_Aint) &s;
    MPI_Type_struct(7, lengths, disps, types, &splInDatatype);
    MPI_Type_commit(&splInDatatype);
  }

  /* jobDatatype */
  /*{
    Job j;
    int lengths[5] = { 1, 1, 2, MAX_NAMELEN, MAX_NAMELEN };
    MPI_Aint disps[5];
    MPI_Datatype types[5];

    types[0] = codecInitInDatatype;
    types[1] = splInDatatype;
    types[2] = MPI_INT;
    types[3] = MPI_CHAR;
    types[4] = MPI_CHAR;

    disps[0] = (MPI_Aint) &j.sCodec;
    disps[1] = (MPI_Aint) &j.sInput;ma
    disps[2] = (MPI_Aint) &j.bDeleteSource;
    disps[3] = (MPI_Aint) &j.outputFilename;
    disps[4] = (MPI_Aint) &j.sourceFilename;
    for (i = 0; i < 5; i++)
      disps[i] -= (MPI_Aint) &j;
    MPI_Type_struct(5, lengths, disps, types, &jobDatatype);
    MPI_Type_commit(&jobDatatype);
  }*/
  {
    Job j;
    int lengths[6] = { 1, 1, 1, 7, MAX_NAMELEN, MAX_NAMELEN };
    MPI_Aint disps[6];
    MPI_Datatype types[6];

    types[0] = codecInitInDatatype;
    types[1] = splInDatatype;
	types[2] = MPI_INT;
	types[3] = MPI_CHAR;
    types[4] = MPI_CHAR;
    types[5] = MPI_CHAR;

    disps[0] = (MPI_Aint) &j.sCodec;
    disps[1] = (MPI_Aint) &j.sInput;
	disps[2] = (MPI_Aint) &j.psNext;
    disps[3] = (MPI_Aint) &j.bDeleteSource;
    disps[4] = (MPI_Aint) &j.outputFilename;
    disps[5] = (MPI_Aint) &j.sourceFilename;
    for (i = 0; i < 6; i++)
      disps[i] -= (MPI_Aint) &j;
    MPI_Type_struct(6, lengths, disps, types, &jobDatatype);
    MPI_Type_commit(&jobDatatype);
  }


  /* slaveJobDatatype */
  {
    slave_job j;
    int lengths[2] = { 1, 3 } ;
    MPI_Aint disps[2];
    MPI_Datatype types[2];

    types[0] = jobDatatype;
    types[1] = MPI_INT;

    disps[0] = (MPI_Aint) &j.theJob;
    disps[1] = (MPI_Aint) &j.weAreDone;
    for (i = 0; i < 2; i++)
      disps[i] -= (MPI_Aint) &j;
    MPI_Type_struct(2, lengths, disps, types, &slaveJobDatatype);
    MPI_Type_commit(&slaveJobDatatype);
  }

  /* slaveReturnDatatype */
  {
    MPI_Type_vector(2, 1, 1, MPI_INT, &slaveReturnDatatype);
    MPI_Type_commit(&slaveReturnDatatype);
  }
}


/*____ parallel_startup() _____________________________________________________*/

void
parallel_startup(int *argc, char **argv[], int *num_available_slaves)
{
  MPI_Comm parent_comm = MPI_COMM_NULL;
  int rank, size, am_slave = 0;

  MPI_Init(argc, argv);

  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#if HAVE_MPI_COMM_GET_PARENT
  MPI_Comm_get_parent(&parent_comm);
#endif

#if HAVE_MPI_COMM_SPAWN && WANT_MPI_SPAWN_UNIVERSE
  /* If we are a singleton, find out how many slaves we can spawn from
      MPI_UNIVERSE_SIZE (the actual spawn is later -- we have to parse
      the command line args first to see if the user wants to override
      MPI_UNIVERSE_SIZE).  We never only spawn 1 child, so the entry
      condition into this logic (size == 1) is sufficient. */

  if (size == 1) {
    int *universe_size, found;

    MPI_Attr_get(MPI_COMM_WORLD, MPI_UNIVERSE_SIZE, &universe_size, &found);
    if (found == 1 && *universe_size > 1)
      *num_available_slaves = *universe_size;
    else
      *num_available_slaves = size - 1;
  } else
    *num_available_slaves = size - 1;
#else
  *num_available_slaves = size - 1;
#endif

  /* Sanity check to make logic below easier */

  if (*num_available_slaves < 2)
    *num_available_slaves = 0;

  /* We're a slave if our parent communicator != MPI_COMM_NULL, or if
     we're not rank 0 */

  if (parent_comm != MPI_COMM_NULL || rank != 0)
    am_slave = 1;

  /* If this is a parallel job, do some startup stuff */

  if (*num_available_slaves > 0) {
    make_mpi_datatypes();

    /* If we're not the master, jump to the parallel_slave() routine
       before we do all the startup stuff below (i.e., only the master
       should do the rest of main()). */

    if (am_slave) {
      wantedQuit = TRUE;
      wantedQuiet = TRUE;

      /* For simplicity, this function will not return */

      parallel_slave(parent_comm);
    }
  }
}

/*____ parallel_spawn() _______________________________________________________*/

void
parallel_spawn(char *argv0, int num_available_slaves, int num_desired_slaves,
	       int *num_slaves, FILE *textout, MPI_Comm *slave_comm)
{
  int rank, size;
#if HAVE_MPI_COMM_SPAWN
  char *spawn_argv[2];
  int *errors;
  MPI_Comm spawn_comm, merged_comm;
  
  spawn_argv[0] = argv0;
  spawn_argv[1] = NULL;
#endif

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  /* Figure out how many slaves we're going to use */
  /* If we have more than one rank already, we just use that many
     slaves (minus 1) */
  
  if (size > 1) {
    *num_slaves = size - 1;
    
    /* If the user specified a -slaves with more than one MPI rank,
       issue an error */
    
    if (num_desired_slaves >= 0) {
      fprintf(textout, "Warning: Used \"-slaves\" option, but started more than 1 MPI rank.\n");
      fprintf(textout, "Warning: \"-slaves\" option will be ignored.\n");
    }
    
    /* Now do the comm split to reorder the communicator */
    
    MPI_Comm_split(MPI_COMM_WORLD, 0, size, slave_comm);
#if HAVE_MPI_COMM_SET_NAME
    MPI_Comm_set_name(*slave_comm, "Slave_communicator");
#endif
  } 
  
  /* Otherwise, we have 1 rank (we'll never have 0 ranks), and we need
     to a) figure out how many slaves to use, and b) spawn them (if we
     have MPI_Comm_spawn) */
  
  else {
    
#if !HAVE_MPI_COMM_SPAWN
    if (num_desired_slaves > 0)
      fprintf(textout, "\nWarning: Used \"-slaves\" option, but the implementation of MPI that you\n         are using does not support the automatic spawning of slaves\n         (MPI_Comm_spawn).  Hence, the \"-slaves\" option will be ignored.\n\n");
    num_desired_slaves = 0;
#else
    /* If the user specified -slaves, use that number, as long as it
       is not 1 ('cause that doesn't make sense). */
    
    if (num_desired_slaves >= 2)
      *num_slaves = num_desired_slaves;
    else if (num_desired_slaves == 1) {
      fprintf(textout, "Warning: Used \"-slaves\" option, but specified 1 slave (which doesn't make\n         sense).  \"-slaves\" option will be ignored.\n");
      *num_slaves = num_available_slaves;
    } 

    /* Maybe they didn't want any slaves at all */

    else if (num_desired_slaves == 0)
      *num_slaves = 0;
    
    /* Otherwise, we have a singleton init, and they didn't specify
       the number of slaves to use, so use the max available */

    else
      *num_slaves = num_available_slaves;
  
    /* Now that we have decided how many slaves to use, we need to
       spawn them. */
    
    if (*num_slaves >= 2) {

#if MASTER_OUT
      printf("Singleton: initiating MPI_Comm_spawn on %d ranks (%s)...\n",
	     *num_slaves, spawn_argv[0]);
#endif
      
      /* Don't really need to check the contents of errors here, nor the
	 return code of MPI_Comm_spawn, because the default error action
	 in MPI is to abort. */
      
      errors = malloc(sizeof(int) * *num_slaves);
      MPI_Comm_spawn(spawn_argv[0], spawn_argv, *num_slaves, 
		     MPI_INFO_NULL, 0, MPI_COMM_WORLD, &spawn_comm, errors);
#if MASTER_OUT
      printf("Singleton: Spawned!\n");
#endif
      free(errors);

      /* Merge the intercommunicator down to an intracommunicator,
	 with the master as the higest rank */
      
      MPI_Intercomm_merge(spawn_comm, 1, &merged_comm);
      
      /* Reorder the communicator so that the slaves come first and the
	 master is last */
      
      MPI_Comm_split(merged_comm, 0, size, slave_comm);
#if HAVE_MPI_COMM_SET_NAME
      MPI_Comm_set_name(*slave_comm, "Slave_communicator");
#endif
      
      MPI_Comm_free(&merged_comm);
      MPI_Comm_free(&spawn_comm);
    }
#endif
  } 
}


/*____ parallel_shutdown() ____________________________________________________*/

/* Tidy up MPI before we exit.  Not strictly necessary, but it's good
   programming etiquette. */
void
parallel_shutdown(MPI_Comm *slave_comm)
{
  if (slaveReturnDatatype != MPI_DATATYPE_NULL)
    MPI_Type_free(&slaveReturnDatatype);
  if (slaveJobDatatype != MPI_DATATYPE_NULL)
    MPI_Type_free(&slaveJobDatatype);
  if (jobDatatype != MPI_DATATYPE_NULL)
    MPI_Type_free(&jobDatatype);
  if (splInDatatype != MPI_DATATYPE_NULL)
    MPI_Type_free(&splInDatatype);
  if (codecInitInDatatype != MPI_DATATYPE_NULL)
    MPI_Type_free(&codecInitInDatatype);

  if (*slave_comm != MPI_COMM_WORLD && *slave_comm != MPI_COMM_NULL)
    MPI_Comm_free(slave_comm);

  MPI_Finalize();
}


/*____ parallel_master() ______________________________________________________*/

/* Primitive algorithm to keep faster workers busy */
void
parallel_master(int samplesPerFrame, char *pBuffer, FILE *fp,
		time_t startTimeBatch, time_t startTimeFile,
		double batchSamplesTotal, double *fileSamplesRead,
		double *batchSamplesRead, int num_slaves, 
		Job *jobInfo, MPI_Comm slave_comm)
{
  int nSamples, nSamples2;
  short *in1, *in2, *curIn, *lastIn, *lastFrame = NULL, *curLastFrame, *temp;
  char *slaveBuffer;

  /* Note that redundantFrames must <= initialFrames */

  int i, who;
  int redundantFrames = 4; /*4;*/
  int redundantSamples = redundantFrames * samplesPerFrame;
  int initialFrames = 32; /*32;*/
  int initialSamples = initialFrames * samplesPerFrame;
  int *slave_size, *chunk_index;
  int chunk_read = -1;
  int maxSamples = samplesPerFrame * 256;
  int num_working_slaves = 0;
  int redundantSamples2, initialSamples2, slave_size_current;
  slave_job job;
  slave_return ret;
#if HAVE_MPI_STATUS_IGNORE
  MPI_Status *status = MPI_STATUS_IGNORE;
#else
  MPI_Status bogus_status, *status = &bogus_status;
#endif
  MPI_Status real_status;

#if MASTER_OUT
  fprintf(textout, "PARALLEL RUN: %d workers (%d samplesPerFrame)\n", 
	  num_slaves, samplesPerFrame);
#endif

  /* Set this job's info */

  job.theJob = *jobInfo;
  job.weAreDone = 0;

  /* Start everyone off with samplesPerFrame samples */

  in1 = malloc(sizeof(short) * initialSamples * num_slaves * 2);
  in2 = malloc(sizeof(short) * maxSamples * 2);
  slave_size = malloc(sizeof(int) * num_slaves);
  chunk_index = malloc(sizeof(int) * num_slaves);
  for (i = 0; i < num_slaves; i++) {
    slave_size[i] = initialSamples;
    chunk_index[i] = -1;
  }
  reset_result_queue();

  /* Send out the first batch */

  job.redundantFrames = 0;
  nSamples = readSamples (&psJobQueue->sInput, 
			  initialSamples * num_slaves, in1);
  /*printf("\ninitialSamples * num_slaves = %d \n nSamples = %d \n size = %d \n",initialSamples * num_slaves, nSamples, psJobQueue->sInput.sampleBits/8);*/
  nSamples2 = nSamples;
  redundantSamples2 = redundantSamples;
  initialSamples2 = initialSamples;
  if (psJobQueue->sCodec.mode != 3) {		
	nSamples2 *= 2;
	redundantSamples2 *=2;
	initialSamples2 *=2;
  }
#if MASTER_OUT
  printf("Master read total of %d inital samples (%d slaves * %d initial samples)\n", nSamples, num_slaves, initialSamples);
#endif
  for (chunk_read = 0, i = 0; i < num_slaves && nSamples > 0; 
       chunk_read++, i++) {
    job.numSamples = (nSamples > initialSamples) ? initialSamples2 : nSamples2;
#if MASTER_OUT 
    printf("Master sending initial load to rank %d (num samples %d),\n", i, 
	   job.numSamples);
#endif

    /* Send the header */

	/*printf("Job adress: %p ,job.theJob.sInput.preReadBuffer: %p, job.theJob.sourceFilename %p \n", &job, &job.theJob.sInput.preReadBuffer, &job.theJob.sourceFilename);*/
    MPI_Send(&job, 1, slaveJobDatatype, i, SEND_JOB_TAG, slave_comm);

    /* If this is not the very first one, send some redundant frames */

    if (i > 0) {
#if MASTER_OUT
      printf("Master sending redundant info: %d num samples\n", 
	     redundantSamples);
#endif
      /*MPI_Send(in1 + ((i * initialFrames) - redundantFrames) * 
	       samplesPerFrame, 
	       redundantSamples2, MPI_SHORT, i, 
	       SEND_DATA_TAG, slave_comm);*/
      MPI_Send(in1 + (i * initialSamples2) - redundantSamples2, 
	       redundantSamples2, MPI_SHORT, i, 
	       SEND_DATA_TAG, slave_comm);
	
    }

    /* Now send the real frames */

#define mymin(a, b) ((a) < (b)) ? (a) : (b)

    MPI_Send(in1 + (i * initialSamples2), 
	     mymin(initialSamples2, nSamples2), MPI_SHORT, i, 
	     SEND_DATA_TAG, slave_comm);

    /* Bookkeeping */

    num_working_slaves++;
    chunk_index[i] = chunk_read;
    job.redundantFrames = redundantFrames;
    nSamples -= initialSamples;
	nSamples2 = nSamples;
	if (psJobQueue->sCodec.mode != 3) {			
		nSamples2 *= 2;
		/*job.redundantFrames = redundantFrames * 2;*/
	}
  }

  /* We don't need such a big buffer anymore, but we do need to save
     the last redundantFrames frames */

  lastIn = in2;
  lastFrame = in2;
  memcpy(lastFrame, in1 + (num_slaves * initialSamples2) - redundantSamples2,
	 redundantSamples2);
  free(in1);
  in1 = malloc(sizeof(short) * maxSamples * 2);
  curIn = in1;

  /* JMS fix me */

  nSamples = 1;

  /* Wait for returns */
  /*printf("Checkpoint: Entering parallel loop.\n");*/
  while (num_working_slaves > 0) {
#if MASTER_OUT
    printf("Master waiting to receive from any slave....\n");
#endif
    MPI_Recv(&ret, 1, slaveReturnDatatype, MPI_ANY_SOURCE, 
	     RETURN_JOB_TAG, slave_comm, &real_status);
    who = real_status.MPI_SOURCE;
#if MASTER_OUT
    printf("Master received from slave %d who has %d samples\n", 
	   who, slave_size[who]);
#endif
    slaveBuffer = malloc(sizeof(char) * ret.num_bytes);
    *batchSamplesRead += slave_size[who];
    *fileSamplesRead += slave_size[who];
#if MASTER_OUT 
    printf("Master receiving %d bytes from slave %d\n", ret.num_bytes, who);
#endif
    MPI_Recv(slaveBuffer, ret.num_bytes, MPI_CHAR, who,
	     RETURN_DATA_TAG, slave_comm, status);

    /* Queue up this returned buffer for writing (it may have been
       returned out of order) */

    queue_result(fp, slaveBuffer, &ret, chunk_index[who]);
	  
    /* Do we have any more chunks to give to the now-idle slave? */

    if (nSamples > 0) {
      if (slave_size[who] < maxSamples)
	slave_size[who] *= 2;
      
      /* Read in the next (slave_size[who] - nSamples) samples */

      nSamples = readSamples (&psJobQueue->sInput, slave_size[who],
			       curIn);
		
      if (slave_size[who] > nSamples)
	slave_size[who] = nSamples;

	slave_size_current = slave_size[who];	
	nSamples2 = nSamples;
	redundantSamples2 = redundantSamples;
	job.redundantFrames = redundantFrames;	
	if (psJobQueue->sCodec.mode != 3)  {			
		nSamples2 *= 2;
		redundantSamples2 *=2;
		/*job.redundantFrames *=2;*/
		slave_size_current *=2;
	}		
		
		
      if (nSamples > 0) {
	curLastFrame = curIn + slave_size_current /*slave_size[who]*/ - redundantSamples2;
#if MASTER_OUT 
	printf("Master sending %d samples (slave_size %d, frames %d) to slave %d\n\t(plus %d redundant samples)\n", 
	       nSamples, slave_size[who], nSamples / samplesPerFrame, who, 
	       redundantSamples);
#endif
	job.numSamples = nSamples2;
	MPI_Send(&job, 1, slaveJobDatatype, who, SEND_JOB_TAG, slave_comm);
	MPI_Send(lastFrame, redundantSamples2, MPI_SHORT, 
		 who, SEND_DATA_TAG, slave_comm);
	MPI_Send(curIn, nSamples2, MPI_SHORT, who, SEND_DATA_TAG, slave_comm);
	chunk_index[who] = chunk_read++;
	job.redundantFrames = redundantFrames;
	/* Swap the buffers */
	
	temp = curIn;
	curIn = lastIn;
	lastIn = temp;
	lastFrame = curLastFrame;
	  } else {
	/* No more data for this file. */
	/* Note: if we want to assign redundant work to slaves to
	   [potentially] finish this file quicker, this is the place to
	   do it. */
	num_working_slaves--;
      }
    } else {
      /* No more data for this file. */
      /* Note: if we want to assign redundant work to slaves to
         [potentially] finish this file quicker, this is the place to
         do it. */
      num_working_slaves--;
    }

#if MASTER_OUT 
    printf("Master: We're at %lf of %lf total batch samples\n", 
	   batchSamplesRead, batchSamplesTotal);
    printf("Master: We're at %lf of %d file samples\n", *fileSamplesRead,
	   psJobQueue->sInput.length);
#endif
    if (!wantedQuiet)
      {
	updateProgressIndicator (startTimeBatch, batchSamplesTotal,
				 *batchSamplesRead, startTimeFile,
				 psJobQueue->sInput.length,
				 *fileSamplesRead, time (NULL), progType,
				 psJobQueue);
      }
    /* JMS Put in key check stuff here */
    /* printf("Checkpoint: Working slaves: %i.\n",num_working_slaves);*/
  }
  /*printf("Checkpoint: End of parallel loop.\n");*/
  /* The beauty of this scheme is that by the time we get here, the
     file write queue is guaranteed to be empty -- there is no need to
     drain it. */

  /* Clean up */

  free(chunk_index);
	/*printf("Checkpoint: End of cleanup chunk_index.\n");*/
  free(slave_size);
	/*printf("Checkpoint: End of cleanup slave_size.\n");*/
  free(in2);
	/*printf("Checkpoint: End of cleanup in2.\n");*/
  free(in1);
  /*printf("Checkpoint: End of cleanup.\n");*/
#if MASTER_OUT
  /*printf("\nMaster: all done\n");*/
#endif
}

/*____ parallel_master_finish() _______________________________________________*/

void
parallel_master_finish(int num_slaves, MPI_Comm slave_comm)
{
  if (num_slaves > 0) {
    int i;
    slave_job job;

    ZERO_ME(&job);
    job.weAreDone = 1;

    for (i = 0; i < num_slaves; i++)
      MPI_Send(&job, 1, slaveJobDatatype, i, SEND_JOB_TAG, slave_comm);
  }
}

/*____ parallel_slave() _______________________________________________________*/

void
parallel_slave(MPI_Comm parent_comm)
{
  int i, numChunks, size;
  int rank, master;
  int outSize;
  short *redundantBuffer, *inBuffer;
  char *outBuffer, *bogusBuffer = NULL;
  CodecInitOut *pCodecInfo;
  int samplesPerFrame;
  slave_job job;
  slave_return ret;
  MPI_Comm merged_comm, slave_comm;
#if HAVE_MPI_STATUS_IGNORE
  MPI_Status *status = MPI_STATUS_IGNORE;
#else
  MPI_Status bogus_status, *status = &bogus_status;
#endif

  if (parent_comm != MPI_COMM_NULL) 
    /* Merge the parent communicator down with the slaves */
    MPI_Intercomm_merge(parent_comm, 0, &merged_comm);
  else
    merged_comm = MPI_COMM_WORLD;

  /* Reorder the communicator so that the slaves come first and the
     master is last.  This isn't necessary when we don't spawn, but we
     do it here so that it's always the same, regardless of whether we
     were spawned or not. */

  MPI_Comm_split(merged_comm, 0, 0, &slave_comm);
#if HAVE_MPI_COMM_SET_NAME
  MPI_Comm_set_name(slave_comm, "Slave_communicator");
#endif
  
  if (merged_comm != MPI_COMM_WORLD)
    MPI_Comm_free(&merged_comm);

  MPI_Comm_rank(slave_comm, &rank);
  MPI_Comm_size(slave_comm, &master);
  master--;

  /* Continue until the master tells us to quit */

  while (1) {
#if SLAVE_OUT
      printf("Slave %d waiting to receive from master\n", rank);
#endif
    MPI_Recv(&job, 1, slaveJobDatatype, master, SEND_JOB_TAG, slave_comm, 
	     status);
    if (job.weAreDone == 1) {
#if SLAVE_OUT
      printf("Slave %d got exit message, quitting\n", rank);
#endif
      break;
    }

    pCodecInfo = codecInit(&job.theJob.sCodec);
    samplesPerFrame = pCodecInfo->nSamples;

#if SLAVE_OUT
    printf("Slave %d starting next frameset\n", rank);
#endif
    inBuffer = malloc(sizeof(short) * job.numSamples);
#if SLAVE_OUT
    printf("Slave %d receiving %d samples (%d redundant samples)\n", 
	   rank, job.numSamples, job.redundantFrames * samplesPerFrame);
    fflush(stdout);
#endif
    if (job.redundantFrames > 0) {
      redundantBuffer = malloc(sizeof(short) * 
			       job.redundantFrames * samplesPerFrame);
      MPI_Recv(redundantBuffer, job.redundantFrames * samplesPerFrame, 
	       MPI_SHORT, master, SEND_DATA_TAG, slave_comm, 
	       status);
    }
    MPI_Recv(inBuffer, job.numSamples, MPI_SHORT, master, SEND_DATA_TAG, 
	     slave_comm, status);
#if SLAVE_OUT
    printf("Slave %d received %d samples\n", rank, job.numSamples);
    fflush(stdout);
#endif

    /* There are many hard-coded array sizes in this program, so we'll
       just divide up the inBuffer and process it in parts the size of
       samplesPerFrame.  */
    /* JMS: It *might* be ok to do all of them at once, but I don't
       know.  So let's just get it working first, and try to do them
       all at once later. */
    
    numChunks = job.numSamples / samplesPerFrame;
    if ((job.numSamples % samplesPerFrame) != 0)
      numChunks++;
    outBuffer = malloc(sizeof(char) * pCodecInfo->bufferSize * numChunks);
    if (bogusBuffer == NULL) 
      bogusBuffer = malloc(sizeof(char) * pCodecInfo->bufferSize);

    /* Process the redundant frames first to build up some state in
       the codec.  Ditch the resulting encoded data. */

#if SLAVE_OUT
    printf("Slave %d: processing %d redundant frames\n", 
	   rank, job.redundantFrames);
#endif
    for (i = 0; i < job.redundantFrames; i++)
      codecEncodeChunk(samplesPerFrame, redundantBuffer + 
		       (i * samplesPerFrame), bogusBuffer);

    if (job.redundantFrames > 0) {
#if SLAVE_OUT
      printf("Slave %d FLUSHING REDUNDANT!\n", rank);
#endif
      III_FlushBitstream();
    }


    /* Now process the real frames */

#if SLAVE_OUT 
    printf("Slave %d: processing %d chunks (max outsize: %d)\n", 
	   rank, numChunks, pCodecInfo->bufferSize * numChunks);
#endif
    outSize = 0;
    for (i = 0; i < numChunks; i++) {
      size = job.numSamples - (samplesPerFrame * i);
      if (size > samplesPerFrame)
	size = samplesPerFrame;
#if SLAVE_OUT & 0
      printf("Slave %d processing chunk number %d of size %d\n", rank,
	     i, size);
#endif
      outSize += codecEncodeChunk(size, inBuffer + (i * samplesPerFrame), 
				  outBuffer + outSize);
    }

    /* Close out the codec and get any trailing bits */

    
#if SLAVE_OUT
    printf("Slave %d finishing up codec: got %d bytes\n", rank, outSize);
#endif
    {
      int foo = codecExit(outBuffer + outSize, &ret.overflow_bits);
#if SLAVE_OUT
      printf("Slave %d got %d bytes from exitCodc and %d overflow_bits\n",
	     rank, foo, ret.overflow_bits);
#endif
      if (job.redundantFrames > 0)
	outSize += foo;
    }/*ret.overflow_bits = 0;
	III_FlushBitstream();*/

    /* Send the result back */
    
#if SLAVE_OUT
    printf("Slave %d sending %d data bytes back (%d overflow bits)\n", 
	   rank, outSize, ret.overflow_bits);
#endif
    ret.num_bytes = outSize;
    ret.overflow_bits = 0;
    MPI_Send(&ret, 1, slaveReturnDatatype, master, RETURN_JOB_TAG,
	     slave_comm);
    MPI_Send(outBuffer, outSize, MPI_CHAR, master, RETURN_DATA_TAG, 
	     slave_comm);
    
    if (job.redundantFrames > 0)
      free(redundantBuffer);
    free(outBuffer);
    free(inBuffer);
#if SLAVE_OUT
    printf("Salve %d looping around\n", rank);
#endif
  }

  /* That's it! */

  free(bogusBuffer);
#if SLAVE_OUT
  printf("Slave finalizing\n");
#endif

  parallel_shutdown(&slave_comm);
  exit(0);
}


/*____ reset_result_queue() ___________________________________________________*/

static void
reset_result_queue(void)
{
  last_written = -1;
}


/*____ queue_result() _________________________________________________________*/

/* Called by the master to write results from a slave to the output
   file.  The results must be queued because they will likely be
   returned out of order.  We only write to the output file when we
   can do so sequentially (and queue up out of order results until
   they can be written sequentially) */
static void 
queue_result(FILE *fp, char *data, slave_return *ret, int index)
{
  queue_entry *cur, *tail, *temp;

  /* If this is the next block, write it */

#if MASTER_OUT
  printf("Master queueing index %d (last written %d)\n", index, last_written);
#endif
  if (index == last_written + 1) {
#if MASTER_OUT
    printf("Master writing index %d (%d bytes)\n", index, ret->num_bytes);
#endif
    if (fwrite (data, 1, ret->num_bytes, fp) != ret->num_bytes) {
      fprintf (textout,
	       "ERROR: Couldn't write '%s'! Disc probably full.\n",
	       psJobQueue->outputFilename);
      quit(-1);
    }
    free(data);
    last_written++;
    /* See if we can write any more from the queue */
    /* Linked list seems sub-optimal, but if it's kept in order,
       actually does quite nicely. */

#if MASTER_OUT
    printf("Master: Searching list to see if we can advance writing...\n");
#endif
    for (cur = head; cur != NULL; /* ...below ... */) {
#if MASTER_OUT
      printf("Master: in list, index is %d -- looking for %d\n", cur->index,
	     last_written + 1);
#endif
      if (cur->index == last_written + 1) {
#if MASTER_OUT
	printf("Master writing index %d from list (%d bytes)\n", cur->index,
	       cur->ret.num_bytes);
#endif
	if (fwrite (cur->data, 1, cur->ret.num_bytes, fp) != 
	    cur->ret.num_bytes) {
	  fprintf (textout,
		   "ERROR: Couldn't write '%s'! Disc probably full.\n",
		   psJobQueue->outputFilename);
	  quit(-1);
	}
	last_written++;

	/* We just wrote the entry, so delete it from the list */
	/* Advance cur, too */

	free(cur->data);
	temp = cur;
	if (cur != head) {
	  cur->prev->next = cur->next;
	  cur = cur->next;
	} else {
	  head = cur->next;
	  cur = head;
	}
	free(temp);
      }

      /* If we've past it already, then break */

      else if (cur->index > last_written + 1)
	break;
      else
	cur = cur->next;
    }
  }

  /* Otherwise, this wasn't the next one, so queue it up */

  else {
#if MASTER_OUT
    printf("Master queing up for later write\n");
#endif
    temp = malloc(sizeof(queue_entry));
    temp->data = data;
    temp->ret = *ret;
    temp->index = index;
    temp->next = temp->prev = NULL;

    if (head == NULL) {
      /* The list is empty; this is the first entry */

#if MASTER_OUT
      printf("Master got new list\n");
#endif
      head = temp;
    } else if (head->index > temp->index) {
      /* New head of the list.  Logic could just as easily be put
         below, but this is slightly simpler */

#if MASTER_OUT
      printf("Master got new head\n");
#endif
      head->prev = temp;
      temp->next = head;
      head = temp;
    } else {
      /* Put it in the list in the right place */

#if MASTER_OUT
      printf("Master scanning list\n");
#endif
      for (tail = cur = head; cur != NULL; tail = cur, cur = cur->next) {
#if MASTER_OUT
	printf("Checking index: %d ?<? %d\n", temp->index, cur->index);
#endif
	if (temp->index < cur->index) {
#if MASTER_OUT
	  printf("Found it!\n");
#endif
	  temp->prev = cur->prev;
	  temp->next = cur;

	  temp->prev->next = temp;
	  cur->prev = temp;

	  break;
	}
      }

      /* If we didn't put it in, then it must be the new tail */

      if (cur == NULL) {
#if MASTER_OUT
	printf("Didn't find it -- new tail\n");
#endif
	tail->next = temp;
	temp->prev = tail;
      }
    }
  }
#if MASTER_OUT 
  printf("Master leaving: last written %d\n", last_written);
#endif
}


#endif /* WANT_MPI */
