/*
 * Master processing header file
 *
 * Both serial and parallel
 */

#ifndef _master_h_
#define _master_h_

#include <config.h>

#if WANT_MPI
#include <mpi.h>
#endif

#define SLAVE_OUT 0
#define MASTER_OUT 0

/*____ Structure Definitions __________________________________________________*/

/** Old Job Definition **/
/*
typedef struct JobDef Job;

struct JobDef
{
  CodecInitIn sCodec;
  SplIn sInput;
  Job *psNext;
  int fDeleteSource;
  char outputFilename[MAX_NAMELEN];
  char sourceFilename[MAX_NAMELEN];
};*/

typedef struct JobDef Job;

struct JobDef
{
  CodecInitIn			sCodec;
  SI_Stream			sInput;
  Job					*psNext;
  char				bDeleteSource;
  char				bNoGap;
  char				bConcat;
  char				outputFilename[MAX_NAMELEN];
  char				sourceFilename[MAX_NAMELEN];
};


#if WANT_MPI
/* Structure sent to the slaves describing the chunk that they are
   about to do */
typedef struct _slave_job {
  Job theJob;
  int weAreDone;
  int numSamples;
  int redundantFrames;
} slave_job;

/* Structure returned from the slaves describing the MP3 data that
   they are returning */
typedef struct _slave_return {
  int num_bytes;
  int overflow_bits;
} slave_return;

/* Structure for maintaining out-of-order writes to the destination
   .mp3 file */
typedef struct _queue_entry {
  char *data;
  slave_return ret;
  int index;
  struct _queue_entry *prev, *next;
} queue_entry;

#endif /* WANT_MPI */


/*____ Function Prototypes ____________________________________________________*/

void serial_master(int samplesPerFrame, char *pBuffer, FILE *fp,
		   time_t startTimeBatch, time_t startTimeFile,
		   double batchSamplesTotal, double *fileSamplesRead,
		   double *batchSamplesRead);

#if WANT_MPI
void parallel_startup(int *argc, char **argv[], int *num_available_slaves);
void parallel_spawn(char *argv0, int num_available_slaves, 
		    int num_desired_slaves, int *num_slaves,
		    FILE *textout, MPI_Comm *slave_comm);
void parallel_shutdown(MPI_Comm *slave_comm);
void parallel_master(int samplesPerFrame, char *pBuffer, FILE *fp,
		     time_t startTimeBatch, time_t startTimeFile,
		     double batchSamplesTotal, double *fileSamplesRead,
		     double *batchSamplesRead, int num_slaves, 
		     Job *jobInfo, MPI_Comm slave_comm);
void parallel_master_finish(int num_slaves, MPI_Comm slave_comm);
void parallel_slave(MPI_Comm parent_comm);
#endif /* WANT_MPI */


/*____ Enumerated values ______________________________________________________*/

#if WANT_MPI
enum {
  SEND_JOB_TAG,
  SEND_DATA_TAG,

  RETURN_JOB_TAG,
  RETURN_DATA_TAG,

  MAX_TAG
};
#endif /* WANT_MPI */

/* Handy macro to make it purify clean */

#define ZERO_ME(a) memset((a), 0, sizeof((*a)))

#endif /* _master_h_ */
